import asyncio
from aioprocessing import AioQueue as Queue, AioEvent as Event
from multiprocess import Value
from contextlib import asynccontextmanager
from concurrent.futures import ThreadPoolExecutor
import multiprocess
import time

class Channel:
    """
    An asynchronous communication channel class that uses asyncio with multiprocess support
    to facilitate inter-process communication.

    Attributes:
        sender_count (multiprocess.Value): A shared value to track the number of active senders.
        active (aioprocessing.AioEvent): An event that receivers wait on until at least one sender is active.
        cancel_event (aioprocessing.AioEvent): An event to signal cancellation across processes.
        cancel_water (future) : A running task which waits on self.cancel_event
        queue (aioprocessing.AioQueue): An asynchronous queue for messages with a specified capacity.
    """
    def __init__(self, capacity=100):
        """
        Initializes a new Channel instance with a specified queue capacity.

        Args:
            capacity (int): The maximum number of items the queue can hold.
        """
        self.sender_count = Value('i', 0)
        self.active = Event()
        self.cancel_event = Event()
        self.cancel_waiter = None
        self.queue = Queue(capacity)

    async def put(self, item):
        """
        An asynchronous coroutine to add an item to the queue.

        Args:
            item: The item to add to the queue.
        """
        await self.queue.coro_put(item)

    @asynccontextmanager
    async def open(self):
        """
        An asynchronous context manager to track the number of active senders.

        Usage:
            async with channel.open():
                await channel.put(item)
        """
        try:
            with self.sender_count.get_lock():
                self.sender_count.value += 1
            self.active.set()
            yield self
        finally:
            with self.sender_count.get_lock():
                self.sender_count.value -= 1


    async def wait_for_cancel(self):
        """
        An asynchronous coroutine that blocks until the cancel_event is set.
        Polls the event status with a timeout.
        """
        while not self.cancel_event.is_set():
            try:
                await asyncio.wait_for(self.cancel_event.coro_wait(), timeout=1.0)
            except asyncio.TimeoutError:
                continue

    async def get(self):
        """
        An asynchronous coroutine that retrieves an item from the queue or
        raises a StopAsyncIteration if the channel is signaled for cancellation.

        Returns:
            The retrieved item from the queue.

        Raises:
            StopAsyncIteration: If the cancellation event is flagged.
        """
        futures = [
            asyncio.ensure_future(self.queue.coro_get()), 
            self.cancel_waiter
            #asyncio.ensure_future(self.wait_for_cancel())
        ]
        done, pending = await asyncio.wait(
            futures,
            return_when=asyncio.FIRST_COMPLETED
        )

        # Cancel any pending tasks
        if futures[1].done():
            for task in pending:
                task.cancel()
            await asyncio.sleep(0.01)
            raise StopAsyncIteration
        

        for task in done:
            if task.exception():
                raise task.exception()
            return task.result()

    def __aiter__(self):
        """
        Makes the Channel instance an asynchronous iterable.

        Returns:
            The channel instance itself.
        """
        if self.cancel_waiter is None: 
            self.cancel_waiter=asyncio.ensure_future(self.wait_for_cancel())
        return self

    async def flag_cancellation(self):
        """
        An asynchronous coroutine that sets the cancellation event flag,
        indicating that no more messages should be sent or received.
        """
        loop = asyncio.get_running_loop()
        with ThreadPoolExecutor() as executor:
            await loop.run_in_executor(executor, self.cancel_event.set)

    async def __anext__(self):
        """
        An asynchronous iterator method to retrieve the next item from the channel's queue.

        Returns:
            The next item from the queue.

        Raises:
            StopAsyncIteration: If there are no active senders and the queue is empty.
        """
        while True:
            await self.active.coro_wait()
            if self.sender_count.value == 0 and self.queue.empty():
                await asyncio.sleep(0.01)
                self.cancel_event.set()

                await asyncio.sleep(0.01)
                raise StopAsyncIteration
            else:
                future = asyncio.ensure_future(self.get())

                try:
                    await asyncio.wait_for(future, timeout=1)
                    if future.done():
                        return future.result()
                        break
                except asyncio.TimeoutError:
                    continue


import copy
import networkx as nx
import multiprocess
import asyncio
import inspect
from .component import Component
from .port import InputPort, OutputPort
import traceback
from graphviz import Digraph


class Graph:

    """Summary

    Attributes:
        edges (list): list of 2-tuples (InputPort,OutputPort)
        networkx_graph (networkx.DiGraph): Internal networkx graph used for distributing components among workers
        nodes (dict): mapping of names to components in the graph
        processes (list): list of processes spawned by run() or start() calls
    """

    def __init__(self):
        self.nodes = {}
        self.edges = []
        self.processes = []
        self.networkx_graph = nx.DiGraph()  # Directed graph
        self._local_namespace = inspect.currentframe().f_back.f_locals

    def __setattr__(self, name, value):
        """setting an attribute with a Component will add it to the graph

        Args:
            name (TYPE): Name of attribute being set
            value (TYPE): If this is a Component, it is added to the graph
        """
        if isinstance(value, type):
            if issubclass(value, Component):
                self.node(name, value(name))
        else:
            super().__setattr__(name, value)

    def node(self, name, component):
        """Adds a named component to the graph.

        Args:
            name (str): name of the component
            component (Component): component to add to the graph.
        """
        component.graph = self
        component.name = name
        self.nodes[name] = component
        self.__dict__[name] = component
        self._local_namespace[name] = component

    def edge(self, outport, inport):
        """Connects an outport to an inport.

        Args:
            outport (cascata.OutputPort): output port to connect
            inport (cascata.InputPort): input port to connect
        """
        outport.connect(inport)
        self.edges.append((outport, inport))
        self.networkx_graph.add_edge(outport.component, inport.component)

    def _break_cycles(self):
        """
        Breaks cycles in the graph for BFS coloring.
        """
        try:
            cycle = nx.find_cycle(self.networkx_graph, orientation='original')
            # Removing an edge to break the cycle
            self.networkx_graph.remove_edge(*cycle[0])
            return True
        except nx.NetworkXNoCycle:
            return False

    def shard(self, num_workers):
        """Shards the graph into a set of GraphWorkers. Breaks all cycles in the internal networkx graph and runs a topological sort. Assigns components to a worker based on the sort order.

        Args:
            num_workers (int): Number of worker processes to spawn

        Returns:
            list: List of GraphWorkers.
        """
        while self._break_cycles():
            pass
        # Perform a topological sort on the graph
        topo_sorted_nodes = list(nx.topological_sort(self.networkx_graph))
        color_map = {}

        # Assign workers based on topological order
        for index, node in enumerate(topo_sorted_nodes):
            color_map[node] = index % num_workers

        workers = [GraphWorker() for _ in range(num_workers)]
        for component_name, component in self.nodes.items():
            worker_index = color_map.get(component_name, 0)
            workers[worker_index].components.append(component)
        return workers

    def start(self, num_workers=None):
        """Starts the graph, without joining it.

        Args:
            num_workers (int, optional): If specified, spawns this amount of processes. If unspecified, uses either the cpu count or the total number of nodes, whichever is smaller.
        """
        if num_workers is None:
            num_workers = min(len(self.nodes), multiprocess.cpu_count())
        workers = self.shard(num_workers)
        self.processes = []
        for worker in workers:
            p = multiprocess.Process(target=worker.run)
            p.start()
            self.processes.append(p)

    def join(self):
        """Joins all processes in the graph. Handles KeyboardInterrupts gracefully.
        """
        try:
            for p in self.processes:
                p.join()
        except KeyboardInterrupt:
            print("\nGraph execution cancelled by user.")
        finally:
            for p in self.processes:
                p.terminate()

    def run(self, num_workers=None):
        """Starts the graph and joins it

        Args:
            num_workers (int, optional): If specified, spawns this amount of processes. If unspecified, uses either the cpu count or the total number of nodes, whichever is smaller.
        """
        self.start(num_workers)
        self.join()

    def to_dot(self):
        """Convert this graph into a GraphViz object for visualization purposes.

        Returns:
            graphviz.Digraph: GraphViz graph object
        """
        dot = Digraph(comment='Cascata Graph', graph_attr=dict(
            rankdir='TD'), node_attr=dict(shape='box'))

        for component_name, component in self.nodes.items():
            # Create a label for the component with inports, component name, and outports
            inports = '|'.join(f'<{port.name}> {port.name}' for port in sorted(component.inports, key=lambda x: x.name))
            outports = '|'.join(f'<{port.name}> {port.name}' for port in sorted(component.outports, key=lambda x: x.name))
            label = f'{{{{ {inports} }}|{component_name}@{component.__class__.__name__}|{{ {outports} }}}}'

            # Add the component node to the graph
            dot.node(component_name, label=label, shape='record')

        # Add edges based on self.edges
        for outport, inport in self.edges:
            dot.edge(f'{inport.component.name}:{inport.name}', f'{outport.component.name}:{outport.name}')

        return dot


class GraphWorker:

    """Main worker for Cascata Graphs.
    Runs all run() methods of all components in the worker asynchronously.

    Attributes:
        components (list): List of components whos run methods will be spawned on this process
    """

    def __init__(self):
        self.components = []

    def run(self):
        """Creates 
        """
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        try:
            loop.run_until_complete(asyncio.gather(
                *[comp.run() for comp in self.components]))
        except Exception as e:
            print(f"Exception in GraphWorker: {e}")
            traceback.print_exc()
        finally:
            loop.close()



import asyncio
from collections import OrderedDict
from contextlib import AsyncExitStack
from aioitertools import zip as azip
from copy import deepcopy
from .port import InputPort, OutputPort


class Component:
    def __init__(self, name):
        self.name = name
        self.ports={}
        self.graph=None
        self.inports = set()
        self.outports = set()
        self._vals = OrderedDict()  # Holds references to both inports and outports
        self._runner = None
        self._groups = []
   
    def __repr__(self):
        inports_text = ' | '.join(port.name for port in self.inports)
        outports_text = ' | '.join(port.name for port in self.outports)
        component_text = self.name
        
        # Determining the longest string for centering
        max_length = max(len(inports_text), len(component_text), len(outports_text))
        
        # Centering each line based on the longest string
        inports_text_centered = inports_text.center(max_length)
        component_text_centered = component_text.center(max_length)
        outports_text_centered = outports_text.center(max_length)
        
        return f"{'_'*(max_length+4)}\n| {inports_text_centered} |\n| {component_text_centered} |\n| {outports_text_centered} |\n{'¯'*(max_length+4)}"

    def __getattr__(self,attr):
        return self.ports.get(attr)
    
    def add_inport(self, port):
        self.inports.add(port)
        port.component=self
        self.ports[port.name]=port
        self._vals[port]=None

    def add_outport(self, port):
        self.outports.add(port)
        port.component=self
        self.ports[port.name]=port
        self._vals[port]=port

    def set_runner(self, runner_function):
        self._runner = runner_function

    def clk(self, *ports):
        for port in ports:
            self._groups.append((port,))

    def sync(self, *ports):
        self._groups.append((ports))

    async def _group_listener(self, group, execute_runner):
        if execute_runner:
            async def listener(group):
                async for values in azip(*group):
                    self._vals.update(zip(group, values))
                    await self._runner(*self._vals.values())
        else:
            async def listener(group):
                async for values in azip(*group):
                    self._vals.update(zip(group, values))
        
        await listener(group)

    async def run(self):

        async with AsyncExitStack() as stack:
            await asyncio.gather(*[stack.enter_async_context(port.open()) for port in self.outports])
            # Categorize ports based on initialization values
            initports = {port for port in self.inports if port.initialization_value is not None}
    
            # Initialize ports with initialization values
            for port in initports:
                if isinstance(port.initialization_value, OutputPort):
                    async for value in port:
                        self._vals[port] = value
                        break
                else:
                    self._vals[port] = port.initialization_value

            # Check if all ports are initports
            if initports == self.inports:
                # If all inports are initports, call the runner and exit
                await self._runner(*self._vals.values())
                return

            if not self._groups:
                self._groups.append([port for port in self.inports if port.initialization_value is None])
    
            all_grouped_ports = set().union(*self._groups)
            exec_flags=[True for port in self._groups]
            ungrouped_ports = self.inports-all_grouped_ports-initports
    
            # Add each ungrouped port as a group on its own
            for port in ungrouped_ports:
                self._groups.append({port})
                exec_flags.append(False)
            print(self._groups)
            # Prepare and run all group listeners concurrently
            group_listeners = [self._group_listener(group,flag) for group,flag in zip(self._groups,exec_flags)]
            await asyncio.gather(*group_listeners)


def inport(port_name, capacity=10, default=None):
    def decorator(func):
        if not hasattr(func, '_inports'):
            func._inports = []
        func._inports.append((port_name, capacity, default))
        return func
    return decorator

def outport(port_name):
    def decorator(func):
        if not hasattr(func, '_outports'):
            func._outports = []
        func._outports.append(port_name)
        return func
    return decorator

def component(func):
    class ComponentSubclass(Component):
        def __init__(self, *args, **kwargs):
            super().__init__(func.__name__)
            for args in getattr(func, '_inports', []):
                self.add_inport(InputPort(*args))
            for port_name in getattr(func, '_outports', []):
                self.add_outport(OutputPort(port_name))
            self.set_runner(deepcopy(func))

    ComponentSubclass.__name__ = func.__name__
    return ComponentSubclass

